@@ -14,21 +14,23 @@ import kotlinx.coroutines.awaitAll
 import kotlinx.coroutines.newFixedThreadPoolContext
 import kotlinx.coroutines.runBlocking
 import kotlinx.coroutines.withContext
+import org.apache.commons.io.FileUtils
 import org.apache.commons.io.FilenameUtils
 import org.apache.commons.lang3.exception.ExceptionUtils
 import org.apache.http.entity.ContentType
 import org.jsoup.Jsoup
 import org.jsoup.nodes.Document
 import org.jsoup.nodes.Element
+import java.nio.file.Files
+import java.nio.file.Path
 import java.util.concurrent.ConcurrentHashMap
 import java.util.concurrent.CopyOnWriteArraySet
 import java.util.concurrent.atomic.AtomicInteger
 
-private data class VisitResult(val download: Boolean = false, val children: Set<SpongeUri> = setOf())
-
-private val skipVisitResult = VisitResult()
-private val downloadVisitResult = VisitResult(download = true)
+private const val NS_TO_S = 1_000_000_000.0
+private const val KB_TO_B = 1_000.0
 
+private val skipAction: Action = { _: SpongeUri, _: Set<SpongeUri> -> }
 private val htmlMimeTypes = setOf(ContentType.TEXT_HTML.mimeType, ContentType.APPLICATION_XHTML_XML.mimeType)
 
 private val attributeKeys = mapOf(
@@ -49,12 +51,16 @@ private fun toSpongeUri(element: Element, attributeKey: String): SpongeUri? {
     }
 }
 
+typealias Action = suspend (spongeUri: SpongeUri, parents: Set<SpongeUri>) -> Unit
+
 class Sponge(private val spongeService: SpongeService, private val spongeConfig: SpongeConfig) {
     private val requestContext = newFixedThreadPoolContext(spongeConfig.concurrentRequests, "request")
     private val downloadContext = newFixedThreadPoolContext(spongeConfig.concurrentDownloads, "download")
-    private val visitResults = ConcurrentHashMap<SpongeUri, VisitResult>()
+    private val actions = ConcurrentHashMap<SpongeUri, Action>()
     private val downloadedUris = CopyOnWriteArraySet<SpongeUri>()
     private val visitedCount = AtomicInteger()
+    private val completedDownloadCount = AtomicInteger()
+    private val downloadAction: Action = { spongeUri: SpongeUri, _: Set<SpongeUri> -> download(spongeUri) }
 
     fun execute() = runBlocking { visit() }
 
@@ -62,27 +68,22 @@ class Sponge(private val spongeService: SpongeService, private val spongeConfig:
         if (visitedCount.incrementAndGet() > spongeConfig.maximumUris) return
 
         try {
-            val visitResult = getVisitResult(spongeUri)
+            val action = actions.computeIfAbsent(spongeUri) { createAction(spongeUri) }
 
-            if (visitResult !== skipVisitResult) {
-                downloadOrVisitChildren(visitResult, spongeUri, parents)
-            }
+            action(spongeUri, parents)
         } catch (t: Throwable) {
             System.err.println("⚠ Processing failed for $spongeUri: ${t.rootMessage}")
         }
     }
 
-    private fun getVisitResult(spongeUri: SpongeUri) =
-        visitResults.computeIfAbsent(spongeUri) { createVisitResult(spongeUri) }
-
-    private fun createVisitResult(spongeUri: SpongeUri): VisitResult {
+    private fun createAction(spongeUri: SpongeUri): Action {
         val extension = FilenameUtils.getExtension(spongeUri.path)
-        var visitResult = skipVisitResult
+        var action: Action = skipAction
 
         if (spongeConfig.fileExtensions.contains(extension)) {
-            visitResult = downloadVisitResult
+            action = downloadAction
         } else {
-            val response = spongeService.request(spongeUri)
+            val response = spongeService.request(spongeUri.uri)
             val mimeType = ContentType.parse(response.contentType()).mimeType
 
             if (htmlMimeTypes.contains(mimeType)) {
@@ -90,21 +91,29 @@ class Sponge(private val spongeService: SpongeService, private val spongeConfig:
                 val children = getChildren(spongeUri, document)
 
                 if (children.isNotEmpty()) {
-                    visitResult = VisitResult(children = children)
+                    action = createVisitAction(children)
                 }
             } else if (spongeConfig.mimeTypes.contains(mimeType)) {
-                visitResult = downloadVisitResult
+                action = downloadAction
             }
         }
 
-        return visitResult
+        return action
+    }
+
+    private fun createVisitAction(spongeUris: Set<SpongeUri>): Action {
+        return { spongeUri: SpongeUri, parents: Set<SpongeUri> ->
+            if (parents.size < spongeConfig.maximumDepth) {
+                visit(spongeUris, parents + spongeUri)
+            }
+        }
     }
 
     private fun getChildren(spongeUri: SpongeUri, document: Document): Set<SpongeUri> {
         return attributeKeys.entries
             .asSequence()
             .map { getChildren(spongeUri, document, it.key, it.value) }
-            .flatMap { it }
+            .flatten()
             .toSet()
     }
 
@@ -126,27 +135,41 @@ class Sponge(private val spongeService: SpongeService, private val spongeConfig:
             spongeConfig.includeSubdomains && host.endsWith(spongeConfig.spongeUri.host)
     }
 
-    private suspend fun downloadOrVisitChildren(
-        visitResult: VisitResult,
-        spongeUri: SpongeUri,
-        parents: Set<SpongeUri>
-    ) {
-        if (visitResult.download) {
-            download(spongeUri)
-        } else if (parents.size < spongeConfig.maximumDepth) {
-            visit(visitResult.children, parents + spongeUri)
-        }
-    }
-
-    private suspend fun download(spongeUri: SpongeUri) {
-        if (downloadedUris.add(spongeUri)) {
-            withContext(downloadContext) { spongeService.download(spongeUri) }
-        }
-    }
-
     private suspend fun visit(spongeUris: Set<SpongeUri>, parents: Set<SpongeUri>) {
         spongeUris.minus(parents)
             .map { GlobalScope.async(requestContext) { visit(it, parents) } }
             .awaitAll()
     }
+
+    private suspend fun download(spongeUri: SpongeUri) {
+        if (downloadedUris.add(spongeUri)) {
+            val path = getDownloadPath(spongeUri)
+
+            if (!spongeConfig.overwriteExistingFiles && Files.exists(path)) {
+                println("∃ $path")
+            } else {
+                withContext(downloadContext) {
+                    val spongeDownload = spongeService.download(spongeUri.uri, path)
+
+                    printSpongeDownload(path, spongeDownload)
+                }
+            }
+        }
+    }
+
+    private fun getDownloadPath(spongeUri: SpongeUri): Path {
+        return spongeConfig.outputDirectory.resolve(spongeUri.host)
+            .resolve(FilenameUtils.getPath(spongeUri.path))
+            .resolve(FilenameUtils.getName(spongeUri.path))
+            .toAbsolutePath()
+    }
+
+    private fun printSpongeDownload(path: Path, spongeDownload: SpongeDownload) {
+        val humanSize = FileUtils.byteCountToDisplaySize(spongeDownload.size)
+        val speed = NS_TO_S * spongeDownload.size / (KB_TO_B * spongeDownload.duration)
+        val humanSpeed = "%.2f kB/s".format(speed)
+        val completedDownloads = "${completedDownloadCount.incrementAndGet()}/${downloadedUris.size}"
+
+        println("↓ $path [$humanSize] [$humanSpeed] [$completedDownloads]")
+    }
 }
