@@ -97,24 +97,20 @@ class SpongeCommand : CliktCommand(name = "sponge", printHelpOnEmptyArgs = true)
         }
 
         try {
-            val spongeServiceConfig = SpongeServiceConfig(
-                outputDirectory,
-                referrer,
-                userAgent,
-                overwriteExistingFiles
-            )
-
+            val spongeServiceConfig = SpongeServiceConfig(referrer, userAgent)
             val spongeService = SpongeService(spongeServiceConfig)
 
             val spongeConfig = SpongeConfig(
                 spongeUri,
+                outputDirectory,
                 mimeTypes.toSet(),
                 fileExtensions.toSet(),
                 maximumDepth,
                 maximumUris,
                 includeSubdomains,
                 concurrentRequests,
-                concurrentDownloads
+                concurrentDownloads,
+                overwriteExistingFiles
             )
 
             Sponge(spongeService, spongeConfig).execute()
