@@ -12,12 +12,14 @@ import io.mockk.every
 import io.mockk.mockk
 import io.mockk.verify
 import org.apache.commons.io.FileUtils
+import org.apache.commons.io.FilenameUtils
 import org.apache.http.entity.ContentType
 import org.jsoup.Connection
 import org.junit.jupiter.api.BeforeEach
 import org.junit.jupiter.api.Test
 import java.io.IOException
 import java.net.URL
+import java.nio.file.Path
 import java.nio.file.Paths
 
 private fun String.toSpongeUri() = SpongeUri(this)
@@ -26,10 +28,11 @@ class SpongeTest {
     private val spongeService = mockk<SpongeService>(relaxed = true)
     private val outputDirectory = Paths.get("testOutput").toAbsolutePath()
     private val fileName = "test.txt"
-    private val imageFileName = "test.png"
+    private val spongeDownload = SpongeDownload(1_000_000, 1_000_000_000)
 
     private val spongeConfig = SpongeConfig(
         "https://test.com".toSpongeUri(),
+        outputDirectory,
         setOf(ContentType.TEXT_PLAIN.mimeType),
         setOf("png")
     )
@@ -40,6 +43,8 @@ class SpongeTest {
     @BeforeEach
     fun beforeEach() {
         FileUtils.deleteDirectory(outputDirectory.toFile())
+
+        every { spongeService.download(any(), any()) } returns spongeDownload
     }
 
     @Test
@@ -48,8 +53,8 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify(exactly = 0) { spongeService.download(any()) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify(exactly = 0) { spongeService.download(any(), any()) }
     }
 
     @Test
@@ -58,8 +63,8 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify(exactly = 0) { spongeService.download(any()) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify(exactly = 0) { spongeService.download(any(), any()) }
     }
 
     @Test
@@ -81,15 +86,72 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
+    }
+
+    @Test
+    fun testDocumentWithLinkAlreadyDownloaded() {
+        val fileUri = "${spongeConfig.spongeUri}/$fileName".toSpongeUri()
+
+        givenDocument(
+            spongeConfig.spongeUri,
+            """
+                    <html>
+                        <body>
+                            <a href="$fileUri" />
+                        </body>
+                    </html>
+                """
+        )
+
+        givenFile(fileUri)
+
+        val downloadPath = getDownloadPath(fileUri)
+
+        FileUtils.touch(downloadPath.toFile())
+
+        executeSponge(spongeConfig)
+
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify(exactly = 0) { spongeService.download(fileUri.uri, downloadPath) }
+    }
+
+    @Test
+    fun testDocumentWithLinkAlreadyDownloadedAndOverwrite() {
+        val config = spongeConfig.copy(overwriteExistingFiles = true)
+        val fileUri = "${spongeConfig.spongeUri}/$fileName".toSpongeUri()
+
+        givenDocument(
+            config.spongeUri,
+            """
+                    <html>
+                        <body>
+                            <a href="$fileUri" />
+                        </body>
+                    </html>
+                """
+        )
+
+        givenFile(fileUri)
+
+        val downloadPath = getDownloadPath(fileUri)
+
+        FileUtils.touch(downloadPath.toFile())
+
+        executeSponge(config)
+
+        verify { spongeService.request(config.spongeUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, downloadPath) }
     }
 
     @Test
     fun testDocumentWithLinkAndImage() {
         val fileUri = "${spongeConfig.spongeUri}/$fileName".toSpongeUri()
-        val imageFileUri = "${spongeConfig.spongeUri}/$imageFileName".toSpongeUri()
+        val imageFileUri = "${spongeConfig.spongeUri}/test.png".toSpongeUri()
 
         givenDocument(
             spongeConfig.spongeUri,
@@ -108,11 +170,11 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
-        verify(exactly = 0) { spongeService.request(imageFileUri) }
-        verify { spongeService.download(imageFileUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
+        verify(exactly = 0) { spongeService.request(imageFileUri.uri) }
+        verify { spongeService.download(imageFileUri.uri, getDownloadPath(imageFileUri)) }
     }
 
     @Test
@@ -134,9 +196,9 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify(exactly = 0) { spongeService.request(fileUri) }
-        verify(exactly = 0) { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify(exactly = 0) { spongeService.request(fileUri.uri) }
+        verify(exactly = 0) { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
     }
 
     @Test
@@ -158,9 +220,9 @@ class SpongeTest {
 
         executeSponge(spongeConfigWithSubdomains)
 
-        verify { spongeService.request(spongeConfigWithSubdomains.spongeUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfigWithSubdomains.spongeUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
     }
 
     @Test
@@ -182,8 +244,8 @@ class SpongeTest {
 
         executeSponge(spongeConfigWithSubdomains)
 
-        verify { spongeService.request(spongeConfigWithSubdomains.spongeUri) }
-        verify(exactly = 0) { spongeService.request(fileUri) }
+        verify { spongeService.request(spongeConfigWithSubdomains.spongeUri.uri) }
+        verify(exactly = 0) { spongeService.request(fileUri.uri) }
     }
 
     @Test
@@ -217,10 +279,10 @@ class SpongeTest {
 
         executeSponge(spongeConfigWithDepthTwo)
 
-        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri) }
-        verify { spongeService.request(childDocumentUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri.uri) }
+        verify { spongeService.request(childDocumentUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
     }
 
     @Test
@@ -255,10 +317,10 @@ class SpongeTest {
 
         executeSponge(spongeConfigWithDepthTwo)
 
-        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri) }
-        verify { spongeService.request(childDocumentUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri.uri) }
+        verify { spongeService.request(childDocumentUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
     }
 
     @Test
@@ -289,8 +351,8 @@ class SpongeTest {
 
         executeSponge(spongeConfigWithDepthTwo)
 
-        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri) }
-        verify { spongeService.request(childDocumentUri) }
+        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri.uri) }
+        verify { spongeService.request(childDocumentUri.uri) }
     }
 
     @Test
@@ -308,7 +370,7 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
     }
 
     @Test
@@ -342,10 +404,10 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify { spongeService.request(childDocumentUri) }
-        verify(exactly = 0) { spongeService.request(fileUri) }
-        verify(exactly = 0) { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify { spongeService.request(childDocumentUri.uri) }
+        verify(exactly = 0) { spongeService.request(fileUri.uri) }
+        verify(exactly = 0) { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
     }
 
     @Test
@@ -365,8 +427,8 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify(exactly = 0) { spongeService.request(childDocumentUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify(exactly = 0) { spongeService.request(childDocumentUri.uri) }
     }
 
     @Test
@@ -386,7 +448,7 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
         verify { spongeService.request(any()) }
     }
 
@@ -414,11 +476,11 @@ class SpongeTest {
 
         executeSponge(spongeConfig)
 
-        verify { spongeService.request(spongeConfig.spongeUri) }
-        verify { spongeService.request(failingFileUri) }
-        verify(exactly = 0) { spongeService.download(failingFileUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfig.spongeUri.uri) }
+        verify { spongeService.request(failingFileUri.uri) }
+        verify(exactly = 0) { spongeService.download(failingFileUri.uri, getDownloadPath(failingFileUri)) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
     }
 
     @Test
@@ -444,12 +506,12 @@ class SpongeTest {
 
         executeSponge(otherConfig)
 
-        verify { spongeService.request(otherConfig.spongeUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
+        verify { spongeService.request(otherConfig.spongeUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
 
-        verify(exactly = 0) { spongeService.request(otherFileUri) }
-        verify(exactly = 0) { spongeService.download(otherFileUri) }
+        verify(exactly = 0) { spongeService.request(otherFileUri.uri) }
+        verify(exactly = 0) { spongeService.download(otherFileUri.uri, getDownloadPath(otherFileUri)) }
     }
 
     @Test
@@ -484,10 +546,10 @@ class SpongeTest {
 
         executeSponge(spongeConfigWithDepthTwo)
 
-        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri) }
-        verify { spongeService.request(childDocumentUri) }
-        verify { spongeService.request(fileUri) }
-        verify { spongeService.download(fileUri) }
+        verify { spongeService.request(spongeConfigWithDepthTwo.spongeUri.uri) }
+        verify { spongeService.request(childDocumentUri.uri) }
+        verify { spongeService.request(fileUri.uri) }
+        verify { spongeService.download(fileUri.uri, getDownloadPath(fileUri)) }
     }
 
     private fun givenDocument(
@@ -501,21 +563,28 @@ class SpongeTest {
         every { response.body() } returns htmlContent
         every { response.url() } returns URL(spongeUri.uri)
 
-        every { spongeService.request(spongeUri) } returns response
+        every { spongeService.request(spongeUri.uri) } returns response
     }
 
     private fun givenFile(spongeUri: SpongeUri, mimeType: String = ContentType.TEXT_PLAIN.mimeType) {
         val response = mockk<Connection.Response>()
 
         every { response.contentType() } returns mimeType
-        every { spongeService.request(spongeUri) } returns response
+        every { spongeService.request(spongeUri.uri) } returns response
     }
 
     private fun givenUriFailsConnection(spongeUri: SpongeUri) {
-        every { spongeService.request(spongeUri) } throws IOException("Error!")
+        every { spongeService.request(spongeUri.uri) } throws IOException("Error!")
     }
 
     private fun executeSponge(spongeConfig: SpongeConfig) {
         Sponge(spongeService, spongeConfig).execute()
     }
+
+    private fun getDownloadPath(spongeUri: SpongeUri): Path {
+        return outputDirectory.resolve(spongeUri.host)
+            .resolve(FilenameUtils.getPath(spongeUri.path))
+            .resolve(FilenameUtils.getName(spongeUri.path))
+            .toAbsolutePath()
+    }
 }
